#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_request.hh"
#include "loggers.hh"

#include "converter.hh"

int sip_codec_request::encode (const LibSip__SIPTypesAndValues::Request& msg, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode: ", (const Base_Type&)msg);
  
  TRACE_INITIALIZE(TRACE_LEVEL7, NULL);
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL, ">>> sip_codec_request::encode\n"));
  
  osip_message_t* sip_message = nullptr; // TODO Use smart pointer
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to initialise internal data structures");
    return -1;
  }

  if (encode_request(msg, sip_message) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to encode Request");
    ::osip_message_free(sip_message);
    sip_message = nullptr;
    return -1;
  }
  
  char* buffer = nullptr;
  size_t length = 0;
  result = ::osip_message_to_str(sip_message, &buffer, &length);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to encode data structures");
    ::osip_message_free(sip_message);
    return -1;
  }
  loggers::get_instance().log("sip_codec_request::encode: Message:%s", buffer);
  data = OCTETSTRING(length, (const unsigned char*)buffer);
  osip_free(buffer);
  ::osip_message_free(sip_message);
  
  loggers::get_instance().log_msg("<<< sip_codec_request::encode: data=", data);
  return result;
}

int sip_codec_request::decode (const OCTETSTRING& data, LibSip__SIPTypesAndValues::Request& msg, params* p_params)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::decode: data=", data);

  TRACE_INITIALIZE(TRACE_LEVEL7, NULL);
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL, ">>> sip_codec_request::decode\n"));
  
  // Sanity checks
  if (data.lengthof() == 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Wrong parameters");
    return -1;
  }

  osip_message_t* sip_message;
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to initialise internal data structures");
    return -1;
  }
  result = ::osip_message_parse(sip_message, (const char*)static_cast<const unsigned char*>(data), data.lengthof());
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to parse SIP message");
    ::osip_message_free(sip_message);
    return -1;
  }

  // Fill RequestLine
  decode_request_line(sip_message, msg);

  // Fill Headers
  decode_request_headers(sip_message, msg);

  // Fill MessageBody
  decode_message_body(sip_message, msg);
  
  // Fill Payload
  decode_payload(sip_message, msg);
  
  ::osip_message_free(sip_message);
  loggers::get_instance().log_msg("sip_codec_request::decode: ", msg);
  
  return 0;
}

int sip_codec_request::encode_request(const LibSip__SIPTypesAndValues::Request& p_request, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request");
  
  encode_request_line(p_request.requestLine(), p_sip_message);
  
  encode_request_headers(p_request.msgHeader(), p_sip_message);
  
  if (p_request.messageBody().is_present()) {
    const LibSip__MessageBodyTypes::MessageBody& m = static_cast<const LibSip__MessageBodyTypes::MessageBody&>(*p_request.messageBody().get_opt_value());
    encode_request_message_body(m, p_sip_message);
  }
  
  if (p_request.payload().is_present()) {
    const LibSip__SIPTypesAndValues::Payload& p = static_cast<const LibSip__SIPTypesAndValues::Payload&>(*p_request.payload().get_opt_value());
    encode_request_payload(p, p_sip_message);
  }

  return 0;
} // End of method encode_request

int sip_codec_request::encode_request_line(const LibSip__SIPTypesAndValues::RequestLine& p_request_line, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request_line");

  // Sip method
  std::string str(LibSip__SIPTypesAndValues::Method::enum_to_str(p_request_line.method()));
  char *p = (char*)osip_malloc(str.length() - 2 + 1); // Will be freed by osip_message_free
  ::strcpy(p, str.substr(0, str.length() - 2).c_str()); // Remove _E
  ::osip_message_set_method(p_sip_message, p);

  // Sip uri
  osip_uri_t *uri;
  if (encode_sip_url(p_request_line.requestUri(), &uri) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_line: Faile to encode SipUrl");
    return -1;
  }
  ::osip_message_set_uri(p_sip_message, uri);

  // SIP version
  p = (char*)osip_malloc(p_request_line.sipVersion().lengthof() + 1);
  ::strcpy(p, (char*)static_cast<const char*>(p_request_line.sipVersion()));
  ::osip_message_set_version(p_sip_message, p);

  return 0;
} // End of method encode_request_line

int sip_codec_request::encode_request_headers(const LibSip__SIPTypesAndValues::MessageHeader& p_msg_header, osip_message_t* p_sip_message)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode_request_headers: ", p_msg_header);
  
  // Encode mandatory fields
  // From
  osip_from_t* from_header = nullptr;
  if (encode_from_header(p_msg_header.fromField(), &from_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode From header");
    return -1;
  }
  char* hvalue;
  ::osip_from_to_str(from_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: From:%s", hvalue);
  ::osip_message_set_from(p_sip_message, hvalue);
  ::osip_from_free(from_header);
  osip_free(hvalue);
  
  // To
  osip_to_t* to_header = nullptr;
  if (encode_to_header(p_msg_header.toField(), &to_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode To header");
    return -1;
  }
  int r = ::osip_to_to_str(to_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: To:'%s'- result:%d", hvalue, r);
  ::osip_message_set_to(p_sip_message, hvalue);
  ::osip_to_free(to_header);
  osip_free(hvalue);

  // Via
  osip_via_t* via_header = nullptr;
  if (encode_via_header(p_msg_header.via(), &via_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Via header");
    return -1;
  }
  ::osip_via_to_str(via_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: Via:%s", hvalue);
  ::osip_message_set_via(p_sip_message, hvalue);
  ::osip_via_free(via_header);
  osip_free(hvalue);

  // Encode Optional fields
  loggers::get_instance().log("sip_codec_request::encode_request_headers: Encode Optional fields");
  
  // Accept
  if (p_msg_header.accept().is_present()) {
    if (encode_accept_header(p_msg_header.accept(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Accept header");
      return -1;
    }
  }
  
  // AcceptContact
  if (p_msg_header.acceptContact().is_present()) {
    if (encode_accept_contact_header(p_msg_header.acceptContact(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode AcceptContact header");
      return -1;
    }
  }
  
  // Allow
  if (p_msg_header.allow().is_present()) {
    if (encode_allow_header(p_msg_header.allow(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Allow header");
      return -1;
    }
  }
  
  // Authorization
  if (p_msg_header.authorization().is_present()) {
    osip_authorization_t* authorization_header = nullptr;
    if (encode_authorization_header(p_msg_header.authorization(), &authorization_header) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Authorization header");
      return -1;
    }
    int result = ::osip_authorization_to_str(authorization_header, &hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: Authorization: %s - %d", hvalue, result);
    result = ::osip_message_set_authorization(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: SIP Authorization: %p - %d", p_sip_message->authorizations, result);
    ::osip_authorization_free(authorization_header);
    osip_free(hvalue);
  }
  
  // CallId
  if (p_msg_header.callId().is_present()) {
    osip_call_id_t* call_id_header;
    if (encode_call_id_header(p_msg_header.callId(), &call_id_header) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Call_Id header");
      return -1;
    }
    ::osip_call_id_to_str(call_id_header, &hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: Call_Id:%s", hvalue);
    int result = ::osip_message_set_call_id(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: SIP Call_Id: %p - %d", p_sip_message->call_id, result);
    ::osip_call_id_free(call_id_header);
    osip_free(hvalue);
  }
  
  // Contact
  if (p_msg_header.contact().is_present()) {
    osip_contact_t* contact_header;
    if (encode_contact_header(p_msg_header.contact(), &contact_header) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Contact header");
      return -1;
    }
    ::osip_contact_to_str(contact_header, &hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: Contact: %s", hvalue);
    int result = ::osip_message_set_contact(p_sip_message, hvalue);
    loggers::get_instance().log("sip_codec_request::encode_request_headers: SIP Contact: %p - %d", p_sip_message->contacts, result);
    ::osip_contact_free(contact_header);
    osip_free(hvalue);
  }
  
  // ContentLength
  osip_content_length_t* content_length_header = nullptr;
  if (encode_content_length_header(p_msg_header.contentLength(), &content_length_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode ContentLength header");
    return -1;
  }
  ::osip_content_length_to_str(content_length_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: ContentLength:%s", hvalue);
  ::osip_message_set_content_length(p_sip_message, hvalue);
  ::osip_content_length_free(content_length_header);
  osip_free(hvalue);
  
  // ContentType
  osip_content_type_t* content_type_header = nullptr;
  if (encode_content_type_header(p_msg_header.contentType(), &content_type_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode ContentType header");
    return -1;
  }
  ::osip_content_type_to_str(content_type_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: ContentType:%s", hvalue);
  ::osip_message_set_content_type(p_sip_message, hvalue);
  ::osip_content_type_free(content_type_header);
  osip_free(hvalue);
  
  // CSeq
  osip_cseq_t* cseq_header = nullptr;
  if (encode_c_seq_header(p_msg_header.cSeq(), &cseq_header) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode CSeq header");
    return -1;
  }
  ::osip_cseq_to_str(cseq_header, &hvalue);
  loggers::get_instance().log("sip_codec_request::encode_request_headers: Cseq:%s", hvalue);
  ::osip_message_set_cseq(p_sip_message, hvalue);
  ::osip_cseq_free(cseq_header);
  osip_free(hvalue);

  // MaxForwards
  if (p_msg_header.maxForwards().is_present()) {
    if (encode_max_forwards_header(p_msg_header.maxForwards(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode MaxForwards header");
      return -1;
    }
  }
  
  // MinSE
  if (p_msg_header.minSE().is_present()) {
    if (encode_min_se_header(p_msg_header.minSE(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode MinSE header");
      return -1;
    }
  }
  
  // Supported
  if (p_msg_header.supported().is_present()) {
    if (encode_supported_header(p_msg_header.supported(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode Supported header");
      return -1;
    }
  }
  
  // PAccessNetworkInfo
  if (p_msg_header.pAccessNetworkInfo().is_present()) {
    if (encode_p_access_network_info_header(p_msg_header.pAccessNetworkInfo(), &p_sip_message) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_request_headers: Faile to encode PAccessNetworkInfo header");
      return -1;
    }
  }
  
  // TODO continue
  const osip_list_t* p = &(p_sip_message->headers);
  unsigned int pos = 0;
  unsigned int size = ::osip_list_size(p);
  loggers::get_instance().log("Unknown header size: %d\n", size);
  while (pos < size) {
    const osip_header_t* header = (const osip_header_t*)osip_list_get(p, pos++);
    loggers::get_instance().log("%p: hname=%s : hvalue=%s\n", header, header->hname, header->hvalue);
  } // End of 'while' statement
  
  return 0;
} // End of method encode_request_headers

int sip_codec_request::encode_request_message_body(const LibSip__MessageBodyTypes::MessageBody& p_message_body, osip_message_t* p_sip_message)
{ // TODO Rename into encode_message_body
  loggers::get_instance().log(">>> sip_codec_request::encode_request_message_body");

  osip_content_length* content_length = ::osip_message_get_content_length(p_sip_message);
  loggers::get_instance().log("sip_codec_request::encode_message_body: content_length= %s", content_length->value);
  if (p_message_body.ischosen(LibSip__MessageBodyTypes::MessageBody::ALT_sdpMessageBody)) {
    const LibSip__SDPTypes::SDP__Message& sdp = p_message_body.sdpMessageBody();
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: sdp: ", sdp);
    
    sdp_message_t* sdp_body;
    ::sdp_message_init(&sdp_body);

    // Encode mandatory fields
    // Version
    INTEGER version = sdp.protocol__version();
    ::sdp_message_v_version_set(sdp_body, (char*)::strdup(std::to_string(static_cast<const int>(version)).c_str()));
    LibSip__SDPTypes::SDP__Origin origin = sdp.origin();
    // Origin
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: origin: ", origin);
    ::sdp_message_o_origin_set(
                               sdp_body,
                               (char*)::strdup(static_cast<const char*>(origin.user__name())),
                               (char*)::strdup(static_cast<const char*>(origin.session__id())),
                               (char*)::strdup(static_cast<const char*>(origin.session__version())),
                               (char*)::strdup(static_cast<const char*>(origin.net__type())),
                               (char*)::strdup(static_cast<const char*>(origin.addr__type())),
                               (char*)::strdup(static_cast<const char*>(origin.addr()))
                               );
    // Session name
    ::sdp_message_s_name_set(sdp_body, (char*)::strdup(static_cast<const char*>(sdp.session__name())));
    // Time, t
    ::sdp_message_s_name_set(sdp_body, (char*)::strdup(static_cast<const char*>(sdp.session__name())));
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: times: ", sdp.times());
    for (int i = 0; i < sdp.times().lengthof(); i++) {
      const LibSip__SDPTypes::SDP__time& time = sdp.times()[i];
      loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: time: ", time);
      ::sdp_message_t_time_descr_add(sdp_body, (char*)::strdup(static_cast<const char*>(time.time__field().start__time())), (char*)::strdup(static_cast<const char*>(time.time__field().stop__time())));
      if (time.time__repeat().is_present()) {
        const LibSip__SDPTypes::SDP__repeat__list& l = static_cast<const OPTIONAL<LibSip__SDPTypes::SDP__repeat__list>&>(*time.time__repeat().get_opt_value());
        loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: l: ", l);
        for (int j = 0; j < l.lengthof(); j++) {
          const LibSip__SDPTypes::SDP__repeat& r = l[j];
          // TODO            r.repeat__interval();
          // TODO            r.active();
          for (int k = 0; k < r.offsets().lengthof(); k++) {
            const LibSip__SDPTypes::SDP__typed__time& s = r.offsets()[k];
            // TODO              ::sdp_message_r_repeat_add(sdp_body, k, char *value);
          }
          // TODO        ::sdp_message_r_repeat_add(sdp_body, int pos_time_descr, char *value)
        } // End of 'for' statement
      }
    } // End of 'for' statement
    
    // Encode optional fields
    // email, e
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: emails: ", sdp.emails());
    if (sdp.emails().is_present()) {
      const LibSip__SDPTypes::SDP__email__list& l = static_cast<const LibSip__SDPTypes::SDP__email__list&>(*sdp.emails().get_opt_value());
      for (int i = 0; i < l.lengthof(); i++) {
        const LibSip__SDPTypes::SDP__contact& e = l[i];
        std::string str;
        if (e.disp__name().is_present()) {
          str = static_cast<const char*>(static_cast<const CHARSTRING&>(*e.disp__name().get_opt_value()));
          str += " ";
        }
        str += static_cast<const char*>(e.addr__or__phone());
        ::sdp_message_e_email_add (sdp_body, ::strdup(str.c_str()));
      } // End of 'for' statement
    }
    // Information, i
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: information: ", sdp.information());
    if (sdp.information().is_present()) {
      ::sdp_message_i_info_set(sdp_body, 0, ::strdup(static_cast<const CHARSTRING&>(*sdp.information().get_opt_value())));
    }
    // Media list, m
    loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: medias: ", sdp.media__list());
    if (sdp.media__list().is_present()) {
      const LibSip__SDPTypes::SDP__media__desc__list& l = static_cast<const LibSip__SDPTypes::SDP__media__desc__list&>(*sdp.media__list().get_opt_value());
      int i = 0;
      for ( ; i < l.lengthof(); i++) {
        const LibSip__SDPTypes::SDP__media__desc m = l[i];
        loggers::get_instance().log_msg("sip_codec_request::encode_request_message_body: media: ", m);
        //::sdp_message_m_media_add(sdp_body, char *media, char *port, char *number_of_port, char *proto);
      } // End of 'for' statement
      ::sdp_message_endof_media(sdp_body, i);
    }
    // TODO To be continued
    
    loggers::get_instance().log("sip_codec_request::encode_message_body: SDP encoded");
    
    char* buff = nullptr;
    ::sdp_message_to_str(sdp_body, &buff);
    if (buff != nullptr) {
      ::osip_message_set_body(p_sip_message, buff, strlen(buff));
      osip_free(buff);
      sdp_message_free(sdp_body);
    }
  }
  
  return 0;
} // End of method encode_request_message_body

int sip_codec_request::encode_request_payload(const LibSip__SIPTypesAndValues::Payload& p_payload, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request_payload");

  return 0;
} // End of method encode_request_payload

void sip_codec_request::decode_request_headers(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_request_headers");
  
  LibSip__SIPTypesAndValues::MessageHeader headers;
  // Decode mandatory fields
  // From
  LibSip__SIPTypesAndValues::From from_header;
  decode_from_header(::osip_message_get_from(p_sip_message), from_header);
  if (from_header.is_value()) {
    headers.fromField() = from_header;
  } else {
    headers.fromField().set_to_omit();
  }
  // To
  LibSip__SIPTypesAndValues::To to_header;
  decode_to_header(::osip_message_get_to(p_sip_message), to_header);
  if (to_header.is_value()) {
    headers.toField() = to_header;
  } else {
    headers.toField().set_to_omit();
  }
  // Via
  LibSip__SIPTypesAndValues::Via via_header;
  decode_via_header(p_sip_message, via_header);
  if (via_header.is_value()) {
    headers.via() = via_header;
  } else {
    headers.via().set_to_omit();
  }

  // Decode Optional fields
  LibSip__SIPTypesAndValues::Accept accept_header;
  decode_accept_header(p_sip_message, accept_header);
  if (accept_header.is_value()) {
    headers.accept() = accept_header;
  } else {
    headers.accept().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptContact accept_contact_header;
  decode_accept_contact_header(p_sip_message, accept_contact_header);
  if (accept_contact_header.is_value()) {
    headers.acceptContact() = accept_contact_header;
  } else {
    headers.acceptContact().set_to_omit();
  }
  
  LibSip__SIPTypesAndValues::AcceptEncoding accept_encoding_header;
  decode_accept_encoding_header(p_sip_message, accept_encoding_header);
  if (accept_encoding_header.is_value()) {
    headers.acceptEncoding() = accept_encoding_header;
  } else {
    headers.acceptEncoding().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptLanguage accept_language_header;
  decode_accept_language_header(p_sip_message, accept_language_header);
  if (accept_language_header.is_value()) {
    headers.acceptLanguage() = accept_language_header;
  } else {
    headers.acceptLanguage().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AlertInfo alert_info_header;
  decode_alert_info_header(p_sip_message, alert_info_header);
  if (alert_info_header.is_value()) {
    headers.alertInfo() = alert_info_header;
  } else {
    headers.alertInfo().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Allow allow_header;
  decode_allow_header(p_sip_message, allow_header);
  if (allow_header.is_value()) {
    headers.allow() = allow_header;
  } else {
    headers.allow().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AllowEvents allow_events_header;
  decode_allow_events_header(p_sip_message, allow_events_header);
  if (allow_events_header.is_value()) {
    headers.allowEvents() = allow_events_header;
  } else {
    headers.allowEvents().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Authorization authorization_header;
  decode_authorization_header(p_sip_message, authorization_header);
  if (authorization_header.is_value()) {
    headers.authorization() = authorization_header;
  } else {
    headers.authorization().set_to_omit();
  }
  
  headers.authenticationInfo().set_to_omit();

  LibSip__SIPTypesAndValues::CallId call_id_header;
  decode_call_id_header(p_sip_message, call_id_header);
  if (call_id_header.is_value()) {
    headers.callId() = call_id_header;
  } else {
    headers.callId().set_to_omit();
  }
  
  headers.callInfo().set_to_omit();

  LibSip__SIPTypesAndValues::Contact contact_header;
  decode_contact_header(p_sip_message, contact_header);
  if (contact_header.is_value()) {
    headers.contact() = contact_header;
  } else {
    headers.contact().set_to_omit();
  }
  
  headers.contentDisposition().set_to_omit();
  headers.contentEncoding().set_to_omit();
  headers.contentLanguage().set_to_omit();
  
  LibSip__SIPTypesAndValues::ContentLength content_length_header;
  decode_content_length_header(p_sip_message, content_length_header);
  if (content_length_header.is_value()) {
    headers.contentLength() = content_length_header;
  } else {
    headers.contentLength().set_to_omit();
  }

  LibSip__SIPTypesAndValues::ContentType content_type_header;
  decode_content_type_header(p_sip_message, content_type_header);
  if (content_type_header.is_value()) {
    headers.contentType() = content_type_header;
  } else {
    headers.contentType().set_to_omit();
  }

  LibSip__SIPTypesAndValues::CSeq c_seq_header;
  decode_c_seq_header(p_sip_message, c_seq_header);
  if (c_seq_header.is_value()) {
    headers.cSeq() = c_seq_header;
  } else {
    headers.cSeq().set_to_omit();
  }
  
  headers.date().set_to_omit();
  headers.errorInfo().set_to_omit();
  headers.event().set_to_omit();
  headers.expires().set_to_omit();
  headers.featureCaps().set_to_omit();
  headers.geolocation().set_to_omit();
  headers.geolocationRouting().set_to_omit();
  headers.historyInfo().set_to_omit();
  headers.infoPackage().set_to_omit();
  headers.inReplyTo().set_to_omit();

  LibSip__SIPTypesAndValues::MaxForwards max_forwards_header;
  decode_max_forwards_header(p_sip_message, max_forwards_header);
  if (max_forwards_header.is_value()) {
    headers.maxForwards() = max_forwards_header;
  } else {
    headers.maxForwards().set_to_omit();
  }

  headers.mimeVersion().set_to_omit();
  headers.minExpires().set_to_omit();
  
  LibSip__SIPTypesAndValues::MinSE min_se_header;
  decode_min_se_header(p_sip_message, min_se_header);
  if (min_se_header.is_value()) {
    headers.minSE() = min_se_header;
  } else {
    headers.minSE().set_to_omit();
  }
  
  headers.organization().set_to_omit();


  // TODO
  headers.pAccessNetworkInfo().set_to_omit();


  
  headers.pAssertedID().set_to_omit();
  headers.pAssertedService().set_to_omit();
  headers.pAssociatedURI().set_to_omit();
  headers.path().set_to_omit();
  headers.pCalledPartyID().set_to_omit();
  headers.pChargingFunctionAddresses().set_to_omit();
  headers.pChargingVector().set_to_omit();
  headers.pEarlyMedia().set_to_omit();
  headers.pMediaAuthorization().set_to_omit();
  headers.pPreferredID().set_to_omit();
  headers.pPreferredService().set_to_omit();
  headers.priority().set_to_omit();
  headers.privacy().set_to_omit();
  headers.proxyAuthenticate().set_to_omit();
  headers.proxyAuthorization().set_to_omit();
  headers.proxyRequire().set_to_omit();
  headers.pVisitedNetworkID().set_to_omit();
  headers.rAck().set_to_omit();
  headers.rSeq().set_to_omit();
  headers.reason().set_to_omit();
  headers.recordRoute().set_to_omit();
  headers.recvInfo().set_to_omit();
  headers.requestDisposition().set_to_omit();
  headers.referredBy().set_to_omit();
  headers.referTo().set_to_omit();
  headers.referSub().set_to_omit();
  headers.replaces().set_to_omit();
  headers.replyTo().set_to_omit();
  headers.require().set_to_omit();
  headers.retryAfter().set_to_omit();
  
  LibSip__SIPTypesAndValues::Route route_header;
  decode_route_header(p_sip_message, route_header);
  if (route_header.is_value()) {
    headers.route() = route_header;
  } else {
    headers.route().set_to_omit();
  }
  
  headers.securityClient().set_to_omit();
  headers.securityServer().set_to_omit();
  headers.securityVerify().set_to_omit();
  headers.server().set_to_omit();
  headers.serviceRoute().set_to_omit();
  
  LibSip__SIPTypesAndValues::SessionExpires session_expires;
  decode_session_expires_header(p_sip_message, session_expires);
  if (session_expires.is_value()) {
    headers.sessionExpires() = session_expires;
  } else {
    headers.sessionExpires().set_to_omit();
  }
  
  headers.sessionId().set_to_omit();
  headers.sipETag().set_to_omit();
  headers.sipIfMatch().set_to_omit();
  headers.subject().set_to_omit();
  headers.subscriptionState().set_to_omit();
  
  LibSip__SIPTypesAndValues::Supported supported_header;
  decode_supported_header(p_sip_message, supported_header);
  if (supported_header.is_value()) {
    headers.supported() = supported_header;
  } else {
    headers.supported().set_to_omit();
  }
  
  headers.timestamp__().set_to_omit();
  headers.unsupported().set_to_omit();
  headers.userToUser().set_to_omit();
  headers.userAgent().set_to_omit();
  headers.warning().set_to_omit();
  headers.wwwAuthenticate().set_to_omit();
  headers.resourcePriority().set_to_omit();
  headers.answerMode().set_to_omit();
  headers.privAnswerMode().set_to_omit();
  headers.targetDialog().set_to_omit();
  headers.pAnswerState().set_to_omit();
  headers.undefinedHeader__List().set_to_omit();
  
  p_request.msgHeader() = headers;

  // List unprocessed headers
  int i = 0;
  osip_header_t* h_;
  while (::osip_message_get_header(p_sip_message, i++, &h_) == 0) {
    loggers::get_instance().warning("sip_codec_request::decode_request_headers: '%s'/'%s' was not processed", h_->hname, h_->hvalue);
  } // End of 'while' statement
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_request_headers: ", p_request);
} // End of method decode_request_headers

void sip_codec_request::decode_message_body(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_message_body");

  osip_content_length* content_length = ::osip_message_get_content_length(p_sip_message);
  loggers::get_instance().log("sip_codec_request::decode_message_body: content_length= %s", content_length->value);
  if (strcmp((const char*)content_length->value, "0") == 0) {
    p_request.messageBody().set_to_omit();
    return;
  }
  // TODO Align the body the content_length

  osip_content_type_t* content_type = ::osip_message_get_content_type(p_sip_message);
  osip_body_t* body;
  int result = ::osip_message_get_body(p_sip_message, 0, &body);
  loggers::get_instance().log("sip_codec_request::decode_message_body: result=%d", result);
  if (result != 0) {
    p_request.messageBody().set_to_omit();
    return;
  }
  if (content_type != nullptr) {
    loggers::get_instance().log("sip_codec_request::decode_message_body: content-type.type: %s", content_type->type);
    loggers::get_instance().log("sip_codec_request::decode_message_body: content-type.subtype: %s", content_type->subtype);
    LibSip__MessageBodyTypes::MessageBody m;
    if (std::string(content_type->subtype).compare("sdp") == 0) {
      LibSip__SDPTypes::SDP__Message sdp_header;
      sdp_message_t* sdp;
      ::sdp_message_init(&sdp);
      if (::sdp_message_parse(sdp, body->body) == 0) {
        // Mandatroy fields
        // Version, v
        char *hvalue = ::sdp_message_v_version_get(sdp);
        if (hvalue != nullptr) {
          sdp_header.protocol__version() = std::stoi(hvalue);
        }
        // Origin, o
        sdp_header.origin() = LibSip__SDPTypes::SDP__Origin(
                                                            ::sdp_message_o_username_get(sdp),
                                                            ::sdp_message_o_sess_version_get(sdp),
                                                            ::sdp_message_o_sess_version_get(sdp),
                                                            ::sdp_message_o_nettype_get(sdp),
                                                            ::sdp_message_o_addrtype_get(sdp),
                                                            ::sdp_message_o_addr_get(sdp)
                                                            );
        // Session name, s
        if ((hvalue = ::sdp_message_s_name_get(sdp)) != nullptr) {
          sdp_header.session__name() = CHARSTRING(hvalue);
        }
        // Time, t
        int i = 0;
        if ((hvalue = ::sdp_message_t_start_time_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__time__list l;
          do {
            LibSip__SDPTypes::SDP__time t;
            LibSip__SDPTypes::SDP__time__field tf;
            tf.start__time() = CHARSTRING(hvalue);
            tf.stop__time() = CHARSTRING(sdp_message_t_stop_time_get(sdp, i));
            t.time__field() = tf;
            t.time__repeat().set_to_omit(); // TODO
            l[i] = t;
            i += 1;
          } while ((hvalue = ::sdp_message_t_start_time_get(sdp, i)) != nullptr);
          sdp_header.times() = l;
        } else {
          loggers::get_instance().warning("sip_codec_request::decode_message_body: No start time in sdp");
        }
        // Optional fields
        // Emails, e
        i = 0;
        if ((hvalue = ::sdp_message_e_email_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__email__list l;
          do {
            LibSip__SDPTypes::SDP__contact e;
            e.addr__or__phone() = CHARSTRING(hvalue);
            e.disp__name().set_to_omit(); // TODO To be refined
            l[i] = e;
            i += 1;
          } while ((hvalue = ::sdp_message_e_email_get(sdp, i)) != nullptr);
          sdp_header.emails() = OPTIONAL<LibSip__SDPTypes::SDP__email__list>(l);
        } else {
          loggers::get_instance().warning("sip_codec_request::decode_message_body: No emails in sdp");
          sdp_header.emails().set_to_omit();
        }
        // Information, i
        i = 0;
        if ((hvalue = ::sdp_message_i_info_get(sdp, i)) != nullptr) {
          sdp_header.information() = OPTIONAL<CHARSTRING>(CHARSTRING(hvalue));
        } else {
          loggers::get_instance().warning("sip_codec_request::decode_message_body: No information in sdp");
          sdp_header.information().set_to_omit();
        }
        // Media, m
        i = 0;
        if ((hvalue = ::sdp_message_m_media_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__media__desc__list l;
          do {
            // Media description
            LibSip__SDPTypes::SDP__media__field f;
            f.media() = CHARSTRING(hvalue);
            LibSip__SDPTypes::SDP__media__port p;
            hvalue = ::sdp_message_m_port_get(sdp, i);
            p.port__number() = std::stoi(hvalue);
            if ((hvalue = ::sdp_message_m_number_of_port_get(sdp, i)) == nullptr) {
              p.num__of__ports().set_to_omit();
            } else {
              p.num__of__ports() = OPTIONAL<INTEGER>(std::stoi(hvalue));
            }
            f.ports() = p;
            f.transport() = CHARSTRING(::sdp_message_m_proto_get(sdp, i));
            LibSip__SDPTypes::SDP__fmt__list fl;
            int j = 0;
            while ((hvalue = ::sdp_message_m_payload_get(sdp, i, j)) != nullptr) {
              fl[j++] = CHARSTRING(hvalue);
            } // End of 'for' statement
            f.fmts() = fl;
            LibSip__SDPTypes::SDP__media__desc m;
            m.media__field() = f;
            // TODO
            // Media attributes
            m.attributes().set_to_omit();
            m.bandwidth().set_to_omit();
            m.information().set_to_omit();
            m.connections().set_to_omit();
            m.key().set_to_omit();
            l[i] = m;
            i += 1;
          } while ((hvalue = ::sdp_message_m_media_get(sdp, i)) != nullptr);
          sdp_header.media__list() = OPTIONAL<LibSip__SDPTypes::SDP__media__desc__list>(l);
        } else {
          loggers::get_instance().warning("sip_codec_request::decode_message_body: No media in sdp");
          sdp_header.media__list().set_to_omit();
        }
        
        // TODO
        loggers::get_instance().warning("sip_codec_request::decode_message_body: Optional SDP not supported yet");
        sdp_header.uri().set_to_omit();
        sdp_header.phone__numbers().set_to_omit();
        sdp_header.connection().set_to_omit();
        // Bandwidth
        //        sdp_bandwidth_t* b = ::sdp_message_bandwidth_get(sdp, i, j);
        sdp_header.bandwidth().set_to_omit();

        sdp_header.timezone__adjustments().set_to_omit();
        sdp_header.key().set_to_omit();

        // Attribute
        //        sdp_attribute_t* a = ::sdp_message_attribute_get(sdp, i, j);
        sdp_header.attributes().set_to_omit();

        // Set message body
        m.sdpMessageBody() = sdp_header;
        loggers::get_instance().log_msg("sip_codec_request::decode_message_body: sdp:", m.sdpMessageBody());
      } else {
        loggers::get_instance().warning("sip_codec_request::decode_message_body: sdp_message_parse failed, fallback to textplain");
        m.textplain() = CHARSTRING(body->body);
      }
    } else {
      loggers::get_instance().warning("sip_codec_request::decode_message_body: Unsuported content-type, fallback to textplain");
      m.textplain() = CHARSTRING(body->body);
    }
    p_request.messageBody() = OPTIONAL<LibSip__MessageBodyTypes::MessageBody>(m);
  } else if ((body != nullptr) && (body->body != nullptr)) {
    LibSip__MessageBodyTypes::MessageBody m;
    m.textplain() = CHARSTRING(body->body);
    p_request.messageBody() = OPTIONAL<LibSip__MessageBodyTypes::MessageBody>(m);
  } else {
    p_request.messageBody().set_to_omit();
  }
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_message_body", p_request);
} // End of method decode_message_body

void sip_codec_request::decode_payload(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_payload");

  p_request.payload().set_to_omit();
  
} // End of method decode_payload

void sip_codec_request::decode_request_line(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_request_line");
  
  LibSip__SIPTypesAndValues::RequestLine request_line;
  std::string str(::osip_message_get_method(p_sip_message));
  str += "_E";
  request_line.method() = LibSip__SIPTypesAndValues::Method(LibSip__SIPTypesAndValues::Method::str_to_enum(str.c_str()));
  loggers::get_instance().log_msg("sip_codec_request::decode_request_line: Method: ", request_line.method());
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_message_get_uri(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_request::decode_request_line: SipUrl: ", uri);
  request_line.requestUri() = uri;
  request_line.sipVersion() = CHARSTRING(::osip_message_get_version(p_sip_message));
  p_request.requestLine() = request_line;
  // FIXME To be continued

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_request_line: ", p_request);
}

